-- Exit codes for system programs
-- Copyright 2013 B. Persson, Bjorn@Rombobeorn.se
--
-- This library is free software: you can redistribute it and/or modify it
-- under the terms of the GNU General Public License version 3, as published
-- by the Free Software Foundation.


with Ada.Command_Line;

package Berkeley_Exit_Codes is

   -- This package defines exit codes corresponding to those defined in
   -- sysexits.h from BSD. It is suggested that milters should use these exit
   -- codes where it makes sense, in addition to those in Ada.Command_Line,
   -- and maybe other ranges for other kinds of errors. The meaning of each
   -- code is approximately described below.

   subtype Exit_Code is Ada.Command_Line.Exit_Status range 64 .. 78;
   -- Exit_Status'First = EX__BASE, Exit_Status'Last = EX__MAX

   Usage_Error         : constant Exit_Code := 64;  -- EX_USAGE
   -- The command was used incorrectly, e.g., with the wrong number of
   -- arguments, a bad flag, a bad syntax in a parameter, or whatever.

   Data_Format_Error   : constant Exit_Code := 65;  -- EX_DATAERR
   -- The input data was incorrect in some way. This should only be used for
   -- users' data and not for system files.

   Cannot_Open_Input   : constant Exit_Code := 66;  -- EX_NOINPUT
   -- An input file (not a system file) did not exist or was not readable.
   -- This could also include errors like "no message" to a mailer (if it cared
   -- to catch it).

   Unknown_User        : constant Exit_Code := 67;  -- EX_NOUSER
   -- The user specified did not exist. This might be used for mail addresses
   -- or remote logins.

   Unknown_Host        : constant Exit_Code := 68;  -- EX_NOHOST
   -- The host specified did not exist. This is used for mail addresses or
   -- network requests.

   Service_Unavailable : constant Exit_Code := 69;  -- EX_UNAVAILABLE
   -- A service is unavailable. This can occur if a support program or file
   -- does not exist. This can also be used as a catchall message when
   -- something you wanted to do doesn't work, but you don't know why.

   Software_Error      : constant Exit_Code := 70;  -- EX_SOFTWARE
   -- An internal software error has been detected. This should be limited to
   -- non-operating system related errors as possible.

   OS_Error            : constant Exit_Code := 71;  -- EX_OSERR
   -- An operating system error has been detected. This is intended to be used
   -- for such things as "cannot fork", "cannot create pipe", or the like. It
   -- includes things like getuid returning a user that does not exist in the
   -- passwd file.

   OS_File_Error       : constant Exit_Code := 72;  -- EX_OSFILE
   -- Some system file (e.g., /etc/passwd, /etc/utmp, etc.) does not exist,
   -- cannot be opened, or has some sort of error (e.g., syntax error).

   Cannot_Create_File  : constant Exit_Code := 73;  -- EX_CANTCREAT
   -- A (user specified) output file cannot be created.

   IO_Error            : constant Exit_Code := 74;  -- EX_IOERR
   -- An error occurred while doing I/O on some file.

   Temporary_Failure   : constant Exit_Code := 75;  -- EX_TEMPFAIL
   -- Temporary failure indicates something that is not really an error. In
   -- Sendmail, this means that a mailer (e.g.) could not create a connection,
   -- and the request should be reattempted later.

   Protocol_Error      : constant Exit_Code := 76;  -- EX_PROTOCOL
   -- The remote system returned something that was "not possible" during a
   -- protocol exchange.

   Permission_Denied   : constant Exit_Code := 77;  -- EX_NOPERM
   -- You did not have sufficient permission to perform the operation. This is
   -- not intended for file system problems, which should use Cannot_Open_Input
   -- or Cannot_Create_File, but rather for higher level permissions.

   Configuration_Error : constant Exit_Code := 78;  -- EX_CONFIG

end Berkeley_Exit_Codes;


-- The descriptions of the exit codes were copied from sysexits.h, which
-- contained the following text:
--
-- Copyright (c) 1987, 1993
--      The Regents of the University of California.  All rights reserved.
--
-- Redistribution and use in source and binary forms, with or without
-- modification, are permitted provided that the following conditions
-- are met:
-- 1. Redistributions of source code must retain the above copyright
--    notice, this list of conditions and the following disclaimer.
-- 2. Redistributions in binary form must reproduce the above copyright
--    notice, this list of conditions and the following disclaimer in the
--    documentation and/or other materials provided with the distribution.
-- 4. Neither the name of the University nor the names of its contributors
--    may be used to endorse or promote products derived from this software
--    without specific prior written permission.
--
-- THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
-- ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
-- IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
-- ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
-- FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
-- DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
-- OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
-- HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
-- LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
-- OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
-- SUCH DAMAGE.
